### Activity Logging Package Specification

#### Version
- **0.1.0**

#### Overview
This package provides a mechanism to track user activities within a Django application, including user custom events, session durations, and CRUD operations. Logs can be viewed, filtered, and exported directly from the Django admin panel.

### Requirements
```bash
Python (>= 3.12)
Django (>= 5.1)
```

---

### Installation and Configuration

1. **Install the Package**

   ```bash
   pip install git+https://git.communitysystemsfoundation.org/scm/dc/py_activity_log.git#egg=activity_logger
   ```

   Alternatively, add the package to your `requirements.txt` file:

   ```text
   git+https://git.communitysystemsfoundation.org/scm/dc/py_activity_log.git#egg=django_activity_logger
   ```

   Then, install it with:

   ```bash
   pip install -r requirements.txt
   ```

2. **Add to `INSTALLED_APPS`**

   In your Django project's `settings.py`, add the following entry:

   ```python
   INSTALLED_APPS = [
       ...
       'aispl.activity_logger',
       'rest_framework',
   ]
   ```

3. **Create and Configure `activity_logger.py`**

   In the base directory of your project (where `manage.py` is located), create a new file called `activity_logger.py`. This file will handle configuration for the activity logging package. Add the following:

   ```python
   # Activity Logger Settings
   ACTIVITY_LOGGER = {
      'DOWNLOAD_ENABLE':False,
      'INCLUDE_MODELS' : ['migrations', 'group', 'permission', 'user']
   }

   MODIFIED_BY = 'modified_by'
   CREATED_BY = 'created_by'
   ```

   The `activity_logger.py` file will be imported in signals.py file of the package, enabling centralized control over activity logging in your application.

4. **Database Migrations**

   After configuring the settings, run the following commands to create the necessary database tables:

   ```bash
   python manage.py migrate
   ```

---

### Features

#### 1. **Activity Tracking**
   - **WHO, WHAT, WHEN**: Records the identity of the user, details of the modification, and a timestamp for each action.
   - **Track All User Activities**: Automatically tracks all user interactions and operations on models.

#### 2. **Configurable Inclusion for Tracking**
   - **Include Models**: Define models to include from activity tracking using the `INCLUDE_MODELS` variable in `settings.py`.

#### 3. **Export Logs**
   - **CSV Export**: Export logs in CSV format for additional analysis or reporting.

#### 4. **Custom Logging Feature**

- **Log Custom Events**: Users can log custom events directly within their application by calling a utility function provided by the package.

- **Usage**: To log a custom event, import and call the `log_activity` function from the `activity_logger.py` file.

- **Function Signature**:
  ```python
  log_activity(
      created_by: Optional[User],
      event_type: str,
      model_name: str,
      operation: str,
      duration: Optional[timedelta] = None,
      previous_value: Optional[str] = None,
      current_value: Optional[str] = None,
      user_agent: Optional[str] = None
  ) -> ActivityLog
  ```

- **Parameters**:
  - **created_by**: The `User` object representing the user responsible for the action (typically `request.user`), or `None` if the action is system-generated.
  - **event_type**: A string describing the type of event (e.g., "custom_event").
  - **model_name**: The name of the model associated with the activity.
  - **operation**: A description of the action performed. This can also include a custom message.
  - **duration**: (Optional) The duration of the user’s session, or time spent on the operation, if applicable.
  - **user_agent**: (Optional) The User Agent string of the client making the request, typically accessible via `request.META.get('HTTP_USER_AGENT')`.
  - **previous_value**: (Optional) A JSON or text representation of the model state before the operation.
  - **current_value**: (Optional) A JSON or text representation of the model state after the operation.

- **Returns**:
  - **ActivityLog**: The function returns the created `ActivityLog` instance for the logged event.

- **Example Usage**:
  ```python
  from activity_logger import logger
  from datetime import timedelta

  def some_view(request):
      # Your view logic here
      logger.log_activity(
          created_by=request.user,
          event_type='custom_event',
          model_name='SomeModel',
          operation='User clicked on XYZ button',
          duration=timedelta(seconds=120),  # Optional: session or operation duration
          user_agent=request.META.get('HTTP_USER_AGENT'),
          previous_value='old_value',
          current_value='new_value'
      )
  ```

### Project Structure

The package will have the following directory structure:

```
django_activity_logger/
    ├── __init__.py                 # Package initialization
    ├── models.py                   # Database models for activity logs
    ├── views.py                    # API views for activity log CRUD operations
    ├── urls.py                     # URL routing for API endpoints
    ├── middleware.py               # Middleware for automatic activity logging
    ├── serializers.py              # Serializers for model-to-JSON conversion
    ├── admin.py                    # Admin interface setup
    ├── apps.py                     # Django app configuration
    ├── migrations/                 # Database migrations
    ├── tests/                      # Unit and integration tests
    └── settings.py                 # Package settings (if needed)
```

---

### API Endpoints(Optional)

The package will provide the following REST API endpoints for managing activity logs:

- **Retrieve Activity Logs**
  - **Endpoint**: `GET /api/activity-logs/`
  - **Query Parameters**: Filter by user, date range, event type, model, etc.
  - **Response**: List of activity logs.

- **Retrieve Specific Log**
  - **Endpoint**: `GET /api/activity-logs/{id}/`
  - **Response**: Detailed information of a specified activity log.

- **Export Logs**
  - **Endpoint**: `GET /api/activity-logs/export/`
  - **Query Parameters**: Format (CSV, JSON, XML) and filters.
  - **Response**: File download of logs in the specified format.

---

### Database Models

#### ActivityLog Model
This model will store all activity data:

- `id`: Auto-incrementing primary key
- `created_at`: DateTimeField (captures event time)
- `created_by`: ForeignKey (refers to the user performing the action)
- `event_type`: CharField (e.g., login, logout, CRUD operation)
- `model_name`: CharField (nullable, for CRUD operations)
- `operation`: CharField (nullable, indicates CRUD type)
- `user_agent`: TextField (nullable)
- `duration`: DurationField (nullable, records session duration)
- `previous_value`:TextField (nullable, to store the previous value)
- `current_value`:TextField (nullable, to store the current value).

---

### Admin Interface

The Django admin interface provides:
- **Log Viewing**: Displays logs with pagination and relevant details.
- **Advanced Filtering**: Filter logs by user, event type, and date range.
- **Export Options**: Export logs to CSV, JSON, or XML.

---

### Testing

- **Unit Tests**: Tests for each logging functionality.
- **Integration Tests**: Validate compatibility with other Django applications.
- **Performance Tests**: Ensure the middleware performs well under high loads.

---

### Security

Sensitive information will be protected by:
- **Excluding Sensitive Data**: Masking specific fields (e.g., passwords).
- **Access Controls**: Ensuring only authorized users can access, view, or export logs.
- **Rate Limiting**: Applying rate limiting to prevent misuse of API endpoints. 

---