from django.db import models
from django.conf import settings

class ActivityLog(models.Model):
    """
    ActivityLog model to track user actions and events across the application.

    This model logs various events performed by users or system processes on specific models 
    in the application. It stores information such as the type of event, affected model, 
    operation performed, and relevant timestamps. This is useful for auditing, monitoring, 
    and debugging purposes.

    Attributes:
        id (AutoField): Primary key for each log entry, automatically incremented.
        created_at (DateTimeField): Timestamp indicating when the log entry was created.
        created_by (ForeignKey): Reference to the user who performed the action, set to null 
            and blank if the action is system-generated.
        event_type (CharField): Description of the event type (e.g., "UPDATE", "DELETE").
        model_name (CharField): Name of the model affected by the event.
        operation (CharField): Specific operation performed (e.g., "create", "update").
        user_agent (TextField): User agent string from the client that performed the action,
            useful for tracking device or browser information.
        duration (DurationField): The duration of the user's session; default is `None`.
        previous_value (TextField): JSON or text representation of the model's state before the change.
        current_value (TextField): JSON or text representation of the model's state after the change.

    Methods:
        __str__(): Returns a string representation of the ActivityLog instance, useful for 
            display in the Django admin or shell.

    Example:
        A log entry might look like:
        `ActivityLog(created_by='John Doe', event_type='UPDATE', model_name='Order', 
        operation='update', created_at='2024-11-08 14:30:00')`
    """
    
    id = models.AutoField(primary_key=True)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, blank=True)
    event_type = models.CharField(max_length=255)
    model_name = models.CharField(max_length=255, blank=True, null=True)
    operation = models.CharField(max_length=255, blank=True, null=True)
    user_agent = models.TextField(blank=True, null=True)
    duration = models.DurationField(blank=True, null=True)
    previous_value = models.TextField(blank=True, null=True)
    current_value = models.TextField(blank=True, null=True)

    def __str__(self):
        """
        String representation of the ActivityLog instance.

        Returns:
            str: A human-readable string summarizing the created_by, event type, 
            affected model, and creation timestamp of the log entry.
        """
        return f"{self.created_by} - {self.event_type} on {self.model_name} at {self.created_at}"
