import json

from django.db import models
from django.db.models.signals import post_save, post_delete, pre_save, m2m_changed
from django.forms.models import model_to_dict
from django.dispatch import receiver
from django.contrib.contenttypes.models import ContentType

from .models import ActivityLog
from .utils import *

# Load the logger configuration on import
load_logger_config()

# Signal to capture previous state before the model is saved
@receiver(pre_save)
def capture_previous_state(sender, instance, **kwargs):
    """
    Signal handler to capture the previous state of a model instance before it is saved.

    This function listens for the `pre_save` signal, which is triggered just before a model instance 
    is saved. If the model instance already exists (i.e., it has a primary key), it captures the state 
    of the instance before the save operation occurs and stores it in the `_previous_value` attribute 
    of the instance. This can be useful for tracking changes or auditing purposes.

    Args:
        sender (Model): The model class that sent the signal.
        instance (Model instance): The actual instance being saved.
        **kwargs: Additional keyword arguments passed by the signal.

    Workflow:
        1. Checks if the instance has a primary key (i.e., if it already exists in the database).
        2. If the instance exists, it retrieves the previous instance from the database using its 
        primary key (`pk`).
        3. Serializes the previous instance to a dictionary and stores it as a JSON string in the 
        `_previous_value` attribute of the instance.
        4. If the instance does not exist (i.e., it's a new instance), it sets `_previous_value` to 
        `None`.

    Example:
        When updating an existing `Product` instance:
            product = Product.objects.get(pk=1)
            product.name = 'Updated Name'
            product.save()  # This will trigger the signal, capturing the previous state.

    Note:
        The `_previous_value` attribute is temporary and only exists for the duration of the save 
        process. It is not persisted to the database unless explicitly saved to a field.
    """
    try:
        model_name = ContentType.objects.get_for_model(sender).model
    except ContentType.DoesNotExist:
        print(f"ContentType does not exist for model: {sender}")
        return
    
    # Skip logging for models not included or the ActivityLog model itself.
    if sender == ActivityLog or model_name not in include_models:
        return
    
    if instance.pk:
        try:
            previous_instance = sender.objects.get(pk=instance.pk)
            instance._previous_value = json.dumps(serialize_instance(previous_instance), default=str)
        except sender.DoesNotExist:
            # If instance does not exist, this is a new entry
            instance._previous_value = None


@receiver(post_save)
def log_model_create_or_update(sender, instance, created, **kwargs):
    """
    Signal handler to log the creation or update of a model instance.

    This function listens for the `post_save` signal, which is triggered after a model instance is 
    saved (either created or updated). If the model being saved is included in settings 
    and is not an `ActivityLog` instance, it records the activity in the `ActivityLog` model.

    Args:
        sender (Model): The model class that sent the signal.
        instance (Model instance): The actual instance being saved.
        created (bool): Indicates if the instance was newly created (`True`) or updated (`False`).
        **kwargs: Additional keyword arguments passed by the signal.

    Workflow:
        1. Determines the model name of the sender instance using Django's ContentType.
        2. Checks if the model is excluded from logging based on settings.
        3. Logs the event type ('CREATE' or 'UPDATE') along with model name, 
           operation, and the instance's string representation in the `ActivityLog`.

    Example:
        When a new `Order` instance is created:
        `log_model_create_or_update(Order, order_instance, created=True)`

    """
    try:
        model_name = ContentType.objects.get_for_model(sender).model
    except ContentType.DoesNotExist:
        print(f"ContentType does not exist for model: {sender}")
        return
    
    # Skip logging for models not included or the ActivityLog model itself.
    if sender == ActivityLog or model_name not in include_models:
        return

    # Detect if this save was a soft delete
    event_type = 'CREATE' if created else 'UPDATE'
    if not created and instance._previous_value:
        # Deserialize _previous_value if it exists
        previous_data = json.loads(instance._previous_value)
        # Check if this is a soft delete (transition from None to set `deleted_at`)
        if 'deleted_at' in previous_data:
            if previous_data.get('deleted_at') is None and instance.deleted_at is not None:
                event_type = 'DELETE'
                operation = f"Soft deleted {model_name} instance"
            else:
                operation = f"{event_type} {model_name} instance"
        else:
            operation = f"{event_type} {model_name} instance"
    else:
        operation = f"{event_type} {model_name} instance"

    # Get the user performing the operation
    user = getattr(instance, updated_by, None) or getattr(instance, created_by, None)
    
    previous_value = getattr(instance, '_previous_value', None)
    current_value = json.dumps(serialize_instance(instance), default=str)
    
    # Log activity for the model instance.
    log_activity(
        created_by=user,
        event_type=event_type,
        model_name=model_name,
        operation=operation,
        previous_value=previous_value,
        current_value=current_value
    )


@receiver(post_delete)
def log_model_delete(sender, instance, **kwargs):
    """
    Signal handler to log the deletion of a model instance.

    This function listens for the `post_delete` signal, triggered after a model instance 
    is deleted. If the model being deleted is included in settings and is not an `ActivityLog` 
    instance, it records the activity in the `ActivityLog` model.

    Args:
        sender (Model): The model class that sent the signal.
        instance (Model instance): The actual instance being deleted.
        **kwargs: Additional keyword arguments passed by the signal.

    Workflow:
        1. Determines the model name of the sender instance using Django's ContentType.
        2. Checks if the model is excluded from logging based on settings.
        3. Logs the deletion event along with model name, operation, 
           and the instance's string representation in the `ActivityLog`.

    Example:
        When an `Order` instance is deleted:
        `log_model_delete(Order, order_instance)`

    """
    try:
        model_name = ContentType.objects.get_for_model(sender).model
    except ContentType.DoesNotExist:
        print(f"ContentType does not exist for model: {sender}")
        return

    # Skip logging for models not included or the ActivityLog model itself.
    if sender == ActivityLog or model_name not in include_models:
        return

    user = getattr(instance, updated_by, None) or getattr(instance, created_by, None)
    
    # Log activity for the model instance.
    log_activity(
        created_by=user,
        event_type='DELETE',
        model_name=model_name,
        operation=f"Deleted {model_name} instance",
        previous_value=str(instance)
    )

@receiver(m2m_changed)
def log_m2m_changes(sender, instance, action, pk_set, **kwargs):
    """
    Logs changes to many-to-many relationships for any model.
    Handles 'pre_add' and 'pre_remove' actions to capture previous and current states.
    """
    # Only process relevant actions
    if action not in ["pre_add", "pre_remove"]:
        print(f"Skipping action {action}.")
        return

    try:
        # Get the model name of the instance
        model_name = ContentType.objects.get_for_model(instance).model
        through_fields = sender._meta.get_fields()

        # Find the correct ManyToManyField on the main model
        m2m_field = None
        for field in instance._meta.get_fields():
            # Ensure the field is a ManyToManyField and has a `remote_field`
            if isinstance(field, models.ManyToManyField) and field.remote_field.through == sender:
                m2m_field = field
                break

        if not m2m_field:
            raise Exception(f"Could not find ManyToManyField for through model {sender}.")

        # Get the related manager dynamically
        related_manager_name = m2m_field.name  # This is the name of the manager (e.g., 'permissions')

        # Use the related manager to get the previous state
        related_manager = getattr(instance, related_manager_name)
        previous_objects = related_manager.all()
        previous_state = {str(obj): obj.pk for obj in previous_objects}  # Use str(obj) for descriptive names

    except Exception as e:
        print(f"Error determining related manager: {e}")
        return

    # Determine the current state based on the action
    current_state = previous_state.copy()
    related_model = m2m_field.related_model

    if action == 'pre_add':
        added_objects = related_model.objects.filter(pk__in=pk_set)
        current_state.update({str(obj): obj.pk for obj in added_objects})  # Add names for added objects
    elif action == 'pre_remove':
        for pk in pk_set:
            related_obj = related_model.objects.get(pk=pk)
            current_state.pop(str(related_obj), None)  # Remove names for removed objects

    # Prepare previous and current data with descriptive strings
    instance_pk = getattr(instance, 'pk', None)
    instance_str = str(instance)

    previous_data = {
        "id": instance_pk,
        "name": instance_str,
        f"{related_manager_name}": list(previous_state.keys()),  # Use names instead of IDs
    }

    current_data = {
        "id": instance_pk,
        "name": instance_str,
        f"{related_manager_name}": list(current_state.keys()),  # Use names instead of IDs
    }

    # Determine the event type
    event_type = "UPDATE"

    # Identify the user performing the action (if provided)
    user = kwargs.get('user', None)

    # Log the activity
    log_activity(
        created_by=user,
        event_type=event_type,
        model_name=model_name,
        operation=f"{event_type} many-to-many relation for {model_name} '{instance_str}'",
        previous_value=str(previous_data),
        current_value=str(current_data),
    )
