import os
import importlib.util
from typing import Optional, Union
from datetime import timedelta

from django.db import models
from django.contrib.auth.models import User
from django.conf import settings
from django.utils.timezone import now

from .models import ActivityLog

def log_activity(
    created_by: Optional[User], 
    event_type: str, 
    model_name: str, 
    operation: str, 
    duration: Optional[timedelta] = None, 
    previous_value: Optional[str] = None, 
    current_value: Optional[str] = None, 
    user_agent: Optional[str] = None
) -> ActivityLog:
    """
    Logs a user or system activity in the ActivityLog model.

    This function creates a new entry in the `ActivityLog` model, recording various details
    about an event, such as the user involved, type of event, model affected, and relevant
    information about the change (previous and current values).

    Args:
        user (Optional[User]): The User object representing the user responsible for the action, 
            or `None` if it's a system action (e.g., an automated process or background task).
        event_type (str): The type of event (e.g., 'CREATE', 'UPDATE', 'DELETE') being logged.
        model_name (str): The name of the model on which the action was performed.
        operation (str): A description of the operation, typically combining event type and model name.
        duration (Optional[timedelta]): The duration of the user's session; default is `None`.
        previous_value (Optional[str]): A JSON or text representation of the model state before the action.
        current_value (Optional[str]): A JSON or text representation of the model state after the action.
        user_agent (Optional[str]): User agent string identifying the client or device making the request.
        
    Returns:
        ActivityLog: The created `ActivityLog` instance for the logged event.

    Example:
        To log an update action on an `Order` instance:
        ```
        log_activity(
            user=order.modified_by,
            event_type='UPDATE',
            model_name='Order',
            operation='UPDATE Order instance',
            previous_value=str(order_before_update),
            current_value=str(order_after_update)
        )
        ```

    """
    return ActivityLog.objects.create(
        created_by=created_by,
        event_type=event_type,
        model_name=model_name,
        operation=operation,
        duration=duration,
        previous_value=previous_value,
        current_value=current_value,
        user_agent=user_agent,
        created_at=now()
    )


def load_logger_config():
    """
    Load logger configuration from activity_logger.py in the project root.

    Attempts to dynamically import activity_logger.py, which defines settings for 
    model activity logging. The configuration includes the models to include in 
    logging (INCLUDE_MODELS), and fields representing the 'created_by' and 
    'updated_by' users.
    """
    project_root = settings.BASE_DIR
    logger_config_path = os.path.join(project_root, 'activity_logger.py')

    if os.path.exists(logger_config_path):
        try:
            spec = importlib.util.spec_from_file_location("logger_config", logger_config_path)
            logger_config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(logger_config)

            # Load configuration variables
            global include_models, created_by, updated_by
            include_models = getattr(logger_config, 'ACTIVITY_LOGGER', {}).get('INCLUDE_MODELS', [])
            created_by = getattr(logger_config, 'CREATED_BY', 'created_by')
            updated_by = getattr(logger_config, 'UPDATED_BY', 'updated_by')

            print("Logger configuration loaded successfully from activity_logger.py")

        except Exception as e:
            print(f"Error loading logger config: {e}")
    else:
        print("Error: activity_logger.py not found in the project root directory.")


# Serialize instance to exclude ManyToManyField data
def serialize_instance(instance):
    """
    Serialize a model instance to a dictionary, excluding ManyToManyField data.
    """
    sensitive_fields = ['password']  # List of sensitive fields to exclude
    fields = {
        field.name: getattr(instance, field.name, None)
        for field in instance._meta.get_fields()
        if not isinstance(field, models.ManyToManyField) and not field.is_relationand and field.name not in sensitive_fields
    }
    return fields